;; Generated package description from buffer-expose.el  -*- no-byte-compile: t -*-
(define-package "buffer-expose" "0.4.3" "Visual buffer switching using a window grid" '((emacs "25") (cl-lib "0.5")) :commit "c4a1c745123b86c15ba7bb4858255b5252e8440a" :authors '(("Clemens Radermacher" . "clemera@posteo.net")) :maintainer '("Clemens Radermacher" . "clemera@posteo.net") :keywords '("convenience") :url "https://github.com/clemera/buffer-expose")
